* ========================================================================= *
* DB4MEMO.PRG
*
* DATE..........: 02/12/89
* AUTHOR........: Phil Kimble & Silvia Flint
*
* PURPOSE.......: Converts Dbase IV files with memo fields back to Dbase III
*                 format.
*
* REMARKS.......: Create a temporary db3 file with the structure from the db4
*                 file.
*
*      Syntax...: db4memo <db4filename> <db3filename>

* ========================================================================= *
parameter infile, dump


* check for invalid parameters passed
if type('infile') = 'U'  .and.  type('dump') = 'U'
   ? "Syntax: db4conv <db4file> <newfile>"
   return
endif


* check if new files have been created
if .not. FILE(dump + '.dbf') .and. .not. FILE(dump + '.dbt')
   ? "ERROR: New file not found..."
   return
endif



CLEAR

* open files
select 0
use (dump)
zap


dbf3_handle = fopen(infile + ".dbf")
if ferror() <> 0
   ? "ERROR: Opening " + infile + ".dbf"
   return
endif


dbt3_handle = fopen(infile + ".dbt")
if ferror() <> 0
   ? "ERROR: Opening " + infile + ".dbt"
   return
endif


* determine record count from header
record_cnt = SPACE(4)
FSEEK(dbf3_handle,4,0)
FREAD(dbf3_handle, @record_cnt, 4)
record_cnt = LTRIM(STR(BIN2L(record_cnt)))
record_cnt = VAL(record_cnt)


* abort if no records to process
if record_cnt < 0
   fclose()
   fclose()
   ? "No records found.."
   quit
endif


* locate beginning data offset
data_offset = SPACE(2)
FSEEK(dbf3_handle,8,0)
FREAD(dbf3_handle, @data_offset, 2)
data_offset = LTRIM(STR(BIN2I(data_offset)))
data_offset = VAL(data_offset)



* locate number of bytes in each record
record_size = SPACE(2)
FSEEK(dbf3_handle,10,0)
FREAD(dbf3_handle, @record_size, 2)
record_size = LTRIM(STR(BIN2W(record_size)))
record_size = VAL(record_size)



* calculate number of fields
fields_total = INT((data_offset - 32) /32)



* set up field descriptor array to test for presence of memos in dbf
DECLARE fldarray[fields_total], fldlengths[fields_total],fldname[fields_total]
AFILL(fldarray,SPACE(32))
AFILL(fldlengths, 0)
AFILL(fldname,SPACE(10))

memo_cnt    = 0
memoloc_pos = 1
memofound   = .f.


FSEEK(dbf3_handle,32,0)
for i = 1 to fields_total
   fldlen      = 0
   array_buf = fldarray[i]


   FREAD(dbf3_handle, @array_buf, 32)
   fldlen = BIN2I(SUBSTR(array_buf,17,1))


   * store each fields length and name to arrays
   fldlengths[i] = fldlen
   fldname[i]    = substr(array_buf,1,10)

   if substr(array_buf,12,1) = 'M'
       memo_cnt    = memo_cnt + 1
       memofound   = .t.
   endif

   memoloc_pos = memoloc_pos + fldlen

next



* exit if no memo fields found in this file.
if .not. memofound
   fclose()
   fclose()
   ? "File does not have any memos.."
   quit
endif

? "Processing Please Wait...."

* array of memo positions in field descriptor array
DECLARE memfldno[memo_cnt], mm[memo_cnt]


* store memofield positions on array
memoloc_pos = 1
cnt = 1
z   = 1
FSEEK(dbf3_handle,32,0)

for i = 1 to fields_total

   fldlen      = 0
   array_buf = fldarray[i]
   FREAD(dbf3_handle, @array_buf, 32)
   fldlen = BIN2I(SUBSTR(array_buf,17,1))

   * check for memo field type
   if substr(array_buf,12,1) = 'M'
       memfldno[cnt] = i
       memoptr  = memoloc_pos
       mm[z]    = memoptr
       z        = z + 1
       cnt      = cnt + 1
   endif


   * add length of field to counter
   memoloc_pos = memoloc_pos + fldlen

next


*======================================================*
* process memo text
*======================================================*

* position pointer to data offset in dbf
FSEEK(dbf3_handle, (data_offset+1), 0)


buffer      = ""
dbf_buffer  = ""
dbt_buffer  = ""
memoloc     = 0
for i = 1 to record_cnt

   * add text to temporary file
   append blank


   * read a record
   dbf_buffer  = FREADSTR(dbf3_handle, record_size)


   * replace fields into new database records
   pos = 1
   for y = 1 to fields_total

       f = fieldname(y)

       do case
       case type("&f") = 'C'
           replace &f with substr(dbf_buffer, pos, fldlengths[y])

       case type("&f") = 'N'
           replace &f with val(substr(dbf_buffer, pos, fldlengths[y]))

       case type("&f") = 'D'
           date = substr(dbf_buffer, pos, fldlengths[y])
           replace &f with ctod(substr(date,5,2)+'/'+substr(date,7,2)+'/'+substr(date,3,2))

       case type("&f") = 'L'
           logval = substr(dbf_buffer, pos, fldlengths[y])
           logval = iif(logval='T', .t., .f. )
           replace &f with logval
       endcase

       pos = pos + fldlengths[y]

   next


   * write the memo text to new file
   for j = 1 to memo_cnt

       memoloc = read_dbf_data(j)

       if memoloc > 0
           buffer  = ""
           buffer  = read_dbt_text( memoloc )
           if .not. empty(buffer)
               m = fieldname(memfldno[j])
               replace &m  with buffer
           endif
       endif
   next

next

FCLOSE(dbf3_handle)
FCLOSE(dbt3_handle)
CLOSE DATABASES

? "Conversion Complete..."
quit



function read_dbf_data
parameter memcount
* read next record
mmloc    = val(substr(dbf_buffer, mm[memcount],10))
return(mmloc)



function read_dbt_text
* read text from next block
parameter blockno
memohdr = space(4)
   * get the number of bytes in current memo string.
   mstr = SPACE(2)
   FSEEK(dbt3_handle, (blockno * 512)+4, 0)
   FREAD(dbt3_handle, @mstr, 2)

   memolen = LTRIM(STR(BIN2I(mstr)))
   memolen = VAL(memolen)

   FSEEK(dbt3_handle,(blockno*512),0)
   FREAD(dbt3_handle,@memohdr,4)

   * Test for "FFFF0800" in first four bytes
   IF ASC(SUBSTR(memohdr,1,1)) = 255 .AND. ASC(SUBSTR(memohdr,2,1)) = 255;
       .AND. ASC(SUBSTR(memohdr,3,1)) = 8 .AND. ASC(SUBSTR(memohdr,4,1)) = 0

   * read string into buffer

      FSEEK(dbt3_handle, (blockno * 512)+8, 0)
      memolen = memolen - 8
   ELSE
      FSEEK(dbt3_handle, (blockno * 512), 0)
   ENDIF

   dbt_buf = ""
   dbt_buf = SPACE(memolen)
   FREAD(dbt3_handle, @dbt_buf, memolen)


return(dbt_buf)
